@@ -30,7 +30,11 @@ import apt_pkg
 import tempfile
 from gettext import gettext as _
 import os
-import subprocess
+
+import gi
+gi.require_version('PackageKitGlib', '1.0')
+from gi.repository import PackageKitGlib as packagekit
+from gi.repository import Gio
 
 from PyQt5 import uic
 from PyQt5.QtCore import *
@@ -49,6 +53,26 @@ from .DialogAdd import DialogAdd
 from .DialogMirror import DialogMirror
 from .CdromProgress import CdromProgress
 
+from UbuntuDrivers import detect
+import sys
+import apt
+from functools import partial
+import aptsources.distro
+import logging
+
+class DetectDriverThread(QThread):
+  #WARNING class to run detect_drivers() in separate thread
+  #so GUI won't freeze.
+  def __init__(self, swprop):
+    QThread.__init__(self)
+    self.swprop = swprop
+
+  def __del__(self):
+    self.wait()
+
+  def run(self):
+    self.swprop.detect_drivers()
+
 class SoftwarePropertiesQtUI(QWidget):
   def __init__(self, datadir):
       QWidget.__init__(self)
@@ -101,6 +125,12 @@ class SoftwarePropertiesQt(SoftwareProperties):
     self.init_isv_sources()
     self.show_isv_sources()
     self.show_cdrom_sources()
+    # Setup and show the Additonal Drivers tab
+    self.init_drivers()
+
+    # Connect to switch-page before setting initial tab. Otherwise the
+    # first switch goes unnoticed.
+    self.userinterface.tabWidget.currentChanged.connect(self.tab_switched)
 
     self.userinterface.checkbutton_source_code.clicked.connect(self.on_checkbutton_source_code_toggled)
     self.userinterface.button_auth_restore.clicked.connect(self.on_restore_clicked)
@@ -117,15 +147,34 @@ class SoftwarePropertiesQt(SoftwareProperties):
     self.userinterface.treeview_sources.itemClicked.connect(self.on_treeview_sources_cursor_changed)
     self.userinterface.treeview_cdroms.itemChanged.connect(self.on_cdrom_source_toggled)
     self.userinterface.treeview2.itemClicked.connect(self.on_treeview_keys_cursor_changed)
-    button_close = self.userinterface.buttonBox.button(QDialogButtonBox.Close)
-    button_close.setIcon(QIcon.fromTheme("dialog-close"))
-    button_revert = self.userinterface.buttonBox.button(QDialogButtonBox.Reset)
-    button_revert.setIcon(QIcon.fromTheme("edit-undo"))
-    button_revert.clicked.connect(self.on_button_revert_clicked)
+
+    self.button_close = self.userinterface.buttonBox.button(QDialogButtonBox.Close)
+    self.button_close.setIcon(QIcon.fromTheme("dialog-close"))
+    self.button_revert = self.userinterface.buttonBox.button(QDialogButtonBox.Reset)
+    self.button_revert.setIcon(QIcon.fromTheme("edit-undo"))
+    self.button_revert.clicked.connect(self.on_button_revert_clicked)
 
     self.init_distro()
     self.show_distro()
 
+    if options and options.open_tab:
+      self.userinterface.tabWidget.setCurrentIndex(int(options.open_tab))
+
+  def tab_switched(self):
+    # On the additional drivers page, don't show the backend revert button.
+    if self.userinterface.tabWidget.currentIndex() == 4: #vbox_drivers is 4
+      self.button_revert.setVisible(False)
+      if not self.detect_called:
+        #WARNING detect_drivers() runs in separate thread
+        #in DetectDriverThread class so GUI won't freeze
+        #after finish show_drivers() has to run in main thread because it updates the GUI
+        self.detect_driver_thread = DetectDriverThread(self)
+        self.detect_driver_thread.finished.connect(self.show_drivers)
+        self.detect_driver_thread.start()
+
+    else:
+      self.button_revert.setVisible(True)
+
   def init_popcon(self):
     """ If popcon is enabled show the statistics tab and an explanation
         corresponding to the used distro """
@@ -136,7 +185,7 @@ class SoftwarePropertiesQt(SoftwareProperties):
       self.userinterface.label_popcon_desc.setText(text)
       self.userinterface.checkbutton_popcon.setChecked(is_helpful)
     else:
-      self.userinterface.tabWidget.removeTab(4)
+      self.userinterface.tabWidget.removeTab(5)
 
   def init_server_chooser(self):
     """ Set up the widgets that allow to choose an alternate download site """
@@ -677,7 +726,7 @@ class SoftwarePropertiesQt(SoftwareProperties):
     if item == None:
         return
     key = item.text(0)
-    if not self.remove_key(key[:8]):
+    if not self.remove_key(key[:16]):
       title = _("Error removing the key")
       text = _("The key you selected could not be removed. "
                "Please report this as a bug.")
@@ -690,6 +739,20 @@ class SoftwarePropertiesQt(SoftwareProperties):
     self.apt_key.update()
     self.show_keys()
 
+  def on_pktask_progress(self, progress, ptype, udata=(None,)):
+    if ptype == packagekit.ProgressType.PERCENTAGE:
+      perc = progress.get_property('percentage')
+      self._pdialog.setValue(perc)
+
+  def on_pktask_finish(self, source, result, udata=(None,)):
+    results = None
+    try:
+        results = self._pktask.generic_finish(result)
+    except Exception as e:
+      QMessageBox.warning(self.userinterface, _("Could not refresh cache"), str(e))
+    self._pdialog.hide()
+    kapp.quit()
+
   def on_close_button(self):
     """Show a dialog that a reload of the channel information is required
        only if there is no parent defined"""
@@ -704,9 +767,34 @@ class SoftwarePropertiesQt(SoftwareProperties):
         messageBox.setText(text)
         messageBox.exec_()
         if (messageBox.clickedButton() == reloadButton):
-                cmd = ["/usr/bin/qapt-batch", "--update"]
-                subprocess.call(cmd)
-    kapp.quit()
+                self._pdialog = QProgressDialog("<big>{}</big>".format(_("Refreshing software cache")),
+                                                "Cancel", 0, 100, self.userinterface)
+                self._pdialog.setWindowTitle(_("Cache Refresh"))
+                self._pdialog.setCancelButton(None)
+                self._pdialog.setAutoClose(False)
+                self._pdialog.setMinimumWidth(210)
+                self._pdialog.setMinimumHeight(60)
+
+                self._pktask = packagekit.Task()
+                self._pdialog.show()
+                self.userinterface.hide()
+                try:
+                    self._pktask.refresh_cache_async (False, # force
+                                                  None,  # GCancellable
+                                                  self.on_pktask_progress,
+                                                  (None,), # user data
+                                                  self.on_pktask_finish,
+                                                  (None,));
+                except Exception as e:
+                    print("Error while requesting cache refresh: {}".format (e))
+        else:
+            # refresh not wanted, quit immediately
+            kapp.quit()
+    else:
+        # no changes, no cache refresh needed.
+        # we can quit.
+        kapp.quit()
+
 
   def on_button_add_cdrom_clicked(self):
     '''Show a dialog that allows to add a repository located on a CDROM
@@ -747,3 +835,453 @@ class SoftwarePropertiesQt(SoftwareProperties):
 
   def run(self):
     kapp.exec_()
+
+  def on_driver_changes_progress(self, progress, ptype, data=None):
+    self.button_driver_revert.setVisible(False)
+    self.button_driver_apply.setVisible(False)
+    self.button_driver_restart.setVisible(False)
+    self.button_driver_cancel.setVisible(True)
+    self.progress_bar.setVisible(True)
+
+    self.userinterface.label_driver_action.setText("Applying changes...")
+    if ptype == packagekit.ProgressType.PERCENTAGE:
+        prog_value = progress.get_property('percentage')
+        self.progress_bar.setValue(prog_value)
+
+  def on_driver_changes_finish(self, source, result, installs_pending):
+    results = None
+    try:
+      results = self.pk_task.generic_finish(result)
+    except Exception as e:
+      self.on_driver_changes_revert()
+      QMessageBox.warning(self.userinterface, _("Error while applying changes"), str(e))
+
+    if not installs_pending:
+      self.progress_bar.setVisible(False)
+      self.clear_changes()
+      self.apt_cache = apt.Cache()
+      self.set_driver_action_status()
+      self.update_label_and_icons_from_status()
+      self.button_driver_revert.setVisible(True)
+      self.button_driver_apply.setVisible(True)
+      self.button_driver_cancel.setVisible(False)
+    #self.scrolled_window_drivers.set_sensitive(True)
+
+  def on_driver_changes_error(self, transaction, error_code, error_details):
+    self.on_driver_changes_revert()
+    self.set_driver_action_status()
+    self.update_label_and_icons_from_status()
+    self.button_driver_revert.setVisible(True)
+    self.button_driver_apply.setVisible(True)
+    self.button_driver_cancel.setVisible(False)
+    #self.scrolled_window_drivers.set_sensitive(True)
+
+  def on_driver_changes_cancellable_changed(self, transaction, cancellable):
+    self.button_driver_cancel.setEnabled(cancellable)
+
+  def on_driver_changes_apply(self, button):
+    self.pk_task = packagekit.Task()
+    button = self.userinterface.sender()
+    installs = []
+    removals = []
+
+    for pkg in self.driver_changes:
+      if pkg.is_installed:
+          removals.append(self.get_package_id(pkg.installed))
+          # The main NVIDIA package is only a metapackage.
+          # We need to collect its dependencies, so that
+          # we can uninstall the driver properly.
+          if 'nvidia' in pkg.shortname:
+            for dep in self.get_dependencies(self.apt_cache, pkg.shortname, 'nvidia'):
+              dep_pkg = self.apt_cache[dep]
+              if dep_pkg.is_installed:
+                removals.append(self.get_package_id(dep_pkg.installed))
+      else:
+        installs.append(self.get_package_id(pkg.candidate))
+
+    self.cancellable = Gio.Cancellable()
+    try:
+      if removals:
+        installs_pending = False
+        if installs:
+            installs_pending = True
+        self.pk_task.remove_packages_async(removals,
+                  False,  # allow deps
+                  True,  # autoremove
+                  self.cancellable,  # cancellable
+                  self.on_driver_changes_progress,
+                  (None, ),  # progress data
+                  self.on_driver_changes_finish,  # callback ready
+                  installs_pending  # callback data
+         )
+      if installs:
+        self.pk_task.install_packages_async(installs,
+              self.cancellable,  # cancellable
+              self.on_driver_changes_progress,
+              (None, ),  # progress data
+              self.on_driver_changes_finish,  # GAsyncReadyCallback
+              False  # ready data
+       )
+      self.button_driver_revert.setEnabled(False)
+      self.button_driver_apply.setEnabled(False)
+      #self.scrolled_window_drivers.set_sensitive(False)
+    except Exception as e:
+      print("Warning: install transaction not completed successfully: {}".format(e))
+
+  def on_driver_changes_revert(self, button_revert=None):
+    # HACK: set all the "Do not use" first; then go through the list of the
+    #       actually selected drivers.
+    for button in self.no_drv:
+      button.setChecked(True)
+
+    for alias in self.orig_selection:
+      button = self.orig_selection[alias]
+      button.setChecked(True)
+
+    self.clear_changes()
+
+    self.button_driver_revert.setEnabled(False)
+    self.button_driver_apply.setEnabled(False)
+
+  def on_driver_changes_cancel(self, button_cancel):
+    self.transaction.cancel()
+    self.clear_changes()
+
+  def on_driver_restart_clicked(self, button_restart):
+    if 'XDG_CURRENT_DESKTOP' in os.environ:
+      desktop = os.environ['XDG_CURRENT_DESKTOP']
+    else:
+      desktop = 'Unknown'
+
+    if (desktop == 'ubuntu:GNOME' and os.path.exists('/usr/bin/gnome-session-quit')):
+      # argument presents a dialog to cancel reboot
+      subprocess.call(['gnome-session-quit', '--reboot'])
+    elif (desktop == 'XFCE' and
+      os.path.exists('/usr/bin/xfce4-session-logout')):
+      subprocess.call(['xfce4-session-logout'])
+    elif (desktop == 'LXDE' and os.path.exists('/usr/bin/lubuntu-logout')):
+      subprocess.call(['lubuntu-logout'])
+    elif (desktop == 'LXQt' and os.path.exists('/usr/bin/lxqt-leave')):
+      subprocess.call(['lxqt-leave'])
+
+  def clear_changes(self):
+    self.orig_selection = {}
+    self.driver_changes = []
+
+  def init_drivers(self):
+    """Additional Drivers tab"""
+    self.button_driver_revert = QPushButton("Revert")
+    self.button_driver_apply = QPushButton("Apply Changes")
+    self.button_driver_cancel = QPushButton("Cancel")
+    self.button_driver_restart = QPushButton("Restart...")
+
+    self.button_driver_revert.clicked.connect(self.on_driver_changes_revert)
+    self.button_driver_apply.clicked.connect(self.on_driver_changes_apply)
+    self.button_driver_cancel.clicked.connect(self.on_driver_changes_cancel)
+    self.button_driver_restart.clicked.connect(self.on_driver_restart_clicked)
+
+    self.button_driver_revert.setEnabled(False)
+    self.button_driver_revert.setVisible(True)
+    self.button_driver_apply.setEnabled(False)
+    self.button_driver_apply.setVisible(True)
+    self.button_driver_cancel.setVisible(False)
+    self.button_driver_restart.setVisible(False)
+
+    #self.userinterface.box_driver_action.addWidget(self.userinterface.label_driver_action)
+    self.userinterface.box_driver_action.addStretch()
+    self.userinterface.box_driver_action.addWidget(self.button_driver_apply)
+    self.userinterface.box_driver_action.addWidget(self.button_driver_revert)
+    self.userinterface.box_driver_action.addWidget(self.button_driver_restart)
+    self.userinterface.box_driver_action.addWidget(self.button_driver_cancel)
+
+    self.label_driver_detail = QLabel("Searching for available drivers...")
+    self.label_driver_detail.setAlignment(Qt.AlignCenter)
+    self.userinterface.box_driver_detail.addWidget(self.label_driver_detail)
+
+    self.progress_bar = QProgressBar()
+
+    self.userinterface.box_driver_action.addWidget(self.progress_bar)
+    self.progress_bar.setVisible(False)
+
+    self.devices = {}
+    self.detect_called = False
+    self.driver_changes = []
+    self.orig_selection = {}
+        # HACK: the case where the selection is actually "Do not use"; is a little
+        #       tricky to implement because you can't check for whether a package is
+        #       installed or any such thing. So let's keep a list of all the
+        #       "Do not use" radios, set those active first, then iterate through
+        #       orig_selection when doing a Reset.
+    self.no_drv = []
+    self.nonfree_drivers = 0
+    self.ui_building = False
+
+  def detect_drivers(self):
+    # WARNING: This is run in a separate thread.
+    self.detect_called = True
+    try:
+      self.apt_cache = apt.Cache()
+      self.devices = detect.system_device_drivers(self.apt_cache)
+    except:
+      # Catch all exceptions and feed them to apport.
+      #GLib.idle_add(self.label_driver_detail.set_text, _("An error occurred while searching for drivers."))
+      self.label_driver_detail.setText("An error occurred while searching for drivers.")
+      # For apport to catch this exception. See
+      # http://bugs.python.org/issue1230540
+      sys.excepthook(*sys.exc_info())
+      return
+  def on_driver_selection_changed(self, modalias, pkg_name=None):
+    button = self.userinterface.sender()
+    #print(modalias)
+    #print(pkg_name)
+
+    if self.ui_building:
+      return
+
+    pkg = None
+    try:
+      if pkg_name:
+        pkg = self.apt_cache[pkg_name]
+        # If the package depends on dkms
+        # we need to install the correct linux metapackage
+        # so that users get the latest headers
+        if 'dkms' in pkg.candidate.record['Depends']:
+          linux_meta = detect.get_linux(self.apt_cache)
+          if (linux_meta and linux_meta not in self.driver_changes):
+            # Install the linux metapackage
+            lmp = self.apt_cache[linux_meta]
+            if not lmp.is_installed:
+              self.driver_changes.append(lmp)
+    except KeyError:
+      pass
+
+    if button.isChecked():
+      if pkg in self.driver_changes:
+        self.driver_changes.remove(pkg)
+
+      if (pkg is not None and modalias in self.orig_selection and button is not self.orig_selection[modalias]):
+        self.driver_changes.append(pkg)
+    else:
+      if pkg in self.driver_changes:
+        self.driver_changes.remove(pkg)
+
+      # for revert; to re-activate the original radio buttons.
+      if modalias not in self.orig_selection:
+          self.orig_selection[modalias] = button
+
+      if (pkg is not None and pkg not in self.driver_changes and pkg.is_installed):
+          self.driver_changes.append(pkg)
+
+    self.button_driver_revert.setEnabled(bool(self.driver_changes))
+    self.button_driver_apply.setEnabled(bool(self.driver_changes))
+
+  def gather_device_data(self, device):
+    '''Get various device data used to build the GUI.
+
+          return a tuple of (overall_status string, icon, drivers dict).
+          the drivers dict is using this form:
+            {"recommended/alternative": {pkg_name: {
+                                                      'selected': True/False
+                                                      'description':
+'description'
+                                                      'builtin': True/False
+                                                    }
+                                         }}
+             "manually_installed": {"manual": {'selected': True, 'description':
+description_string}}
+             "no_driver": {"no_driver": {'selected': True/False, 'description':
+description_string}}
+
+             Please note that either manually_installed and no_driver are set to
+None if not applicable
+             (no_driver isn't present if there are builtins)
+        '''
+
+    possible_overall_status = {
+            'recommended': (_("This device is using the recommended driver."),
+"recommended-driver"),
+            'alternative': (_("This device is using an alternative driver."),
+"other-driver"),
+            'manually_installed': (_("This device is using a manually-installed driver."), "other-driver"),
+            'no_driver': (_("This device is not working."), "disable-device")
+        }
+
+    returned_drivers = {'recommended': {}, 'alternative': {},
+'manually_installed': {}, 'no_driver': {}}
+    have_builtin = False
+    one_selected = False
+    try:
+      if device['manual_install']:
+        returned_drivers['manually_installed'] = {True: {'selected': True,'description': _("Continue using a manually installed driver")}}
+    except KeyError:
+      pass
+
+    for pkg_driver_name in device['drivers']:
+      current_driver = device['drivers'][pkg_driver_name]
+
+      # get general status
+      driver_status = 'alternative'
+      try:
+        if current_driver['recommended'] and current_driver['from_distro']:
+          driver_status = 'recommended'
+      except KeyError:
+        pass
+
+      builtin = False
+      try:
+        if current_driver['builtin']:
+          builtin = True
+          have_builtin = True
+      except KeyError:
+        pass
+
+      try:
+        pkg = self.apt_cache[pkg_driver_name]
+        installed = pkg.is_installed
+        if pkg.candidate is not None:
+          description = _("Using {} from {}").format(pkg.candidate.summary, pkg.shortname)
+        else:
+          description = _("Using {}").format(pkg.shortname)
+      except KeyError:
+        print("WARNING: a driver ({}) doesn't have any available package associated: {}".format(pkg_driver_name, current_driver))
+        continue
+
+      # gather driver description
+      if current_driver['free']:
+        licence = _("open source")
+      else:
+        licence = _("proprietary")
+
+      if driver_status == 'recommended':
+        base_string = _("{base_description} ({licence}, tested)")
+      else:
+        base_string = _("{base_description} ({licence})")
+      description = base_string.format(base_description=description,
+licence=licence)
+
+      selected = False
+      if not builtin and not returned_drivers['manually_installed']:
+        selected = installed
+        if installed:
+          selected = True
+          one_selected = True
+
+      returned_drivers[driver_status].setdefault(pkg_driver_name, {'selected': selected,'description': description,'builtin': builtin})
+
+    # adjust making the needed addition
+    if not have_builtin:
+      selected = False
+      if not one_selected:
+        selected = True
+      returned_drivers["no_driver"] = {True: {'selected': selected, 'description': _("Do not use the device")}}
+    else:
+      # we have a builtin and no selection: builtin is the selected one then
+      if not one_selected:
+        for section in ('recommended', 'alternative'):
+          for pkg_name in returned_drivers[section]:
+            if returned_drivers[section][pkg_name]['builtin']:
+              returned_drivers[section][pkg_name]['selected'] = True
+
+    # compute overall status
+    for section in returned_drivers:
+      for keys in returned_drivers[section]:
+        if returned_drivers[section][keys]['selected']:
+          (overall_status, icon) = possible_overall_status[section]
+
+    return (overall_status, icon, returned_drivers)
+
+  def show_drivers(self):
+    if not self.devices:
+      # No drivers found.
+      self.label_driver_detail.setText("No additional drivers available.")
+      return
+    else:
+      self.label_driver_detail.hide()
+
+    self.option_group = {}
+    self.radio_button = {}
+    self.ui_building = True
+    self.dynamic_device_status = {}
+    for device in sorted(self.devices.keys()):
+      (overall_status, icon, drivers) = self.gather_device_data(self.devices[device])
+
+      driver_status = QLabel()
+      driver_status.setAlignment(Qt.AlignTop)
+      driver_status.setAlignment(Qt.AlignHCenter)
+      pixmap = QIcon.fromTheme(icon).pixmap(QSize(16, 16))
+      driver_status.setPixmap(pixmap)
+
+      device_box = QHBoxLayout()
+      device_box.addWidget(driver_status)
+      device_detail = QVBoxLayout()
+      device_box.addLayout(device_detail,1)#1 for priority over the icon to stretch
+
+      widget = QLabel("{}: {}".format(self.devices[device].get('vendor', _('Unknown')), self.devices[device].get('model', _('Unknown'))))
+      widget.setAlignment(Qt.AlignLeft)
+      device_detail.addWidget(widget)
+      widget = QLabel("<small>{}</small>".format(overall_status))
+      widget.setAlignment(Qt.AlignLeft)
+      #widget.set_use_markup(True)
+      device_detail.addWidget(widget)
+      self.dynamic_device_status[device] = (driver_status, widget)
+
+      self.option_group[device] = None
+      # define the order of introspection
+      for section in ('recommended', 'alternative', 'manually_installed', 'no_driver'):
+        for driver in drivers[section]:
+          self.radio_button = QRadioButton(drivers[section][driver]['description'])
+          if self.option_group[device]:
+            self.option_group[device].addButton(self.radio_button)
+          else:
+            self.option_group[device] = QButtonGroup()
+            self.option_group[device].addButton(self.radio_button)
+
+          device_detail.addWidget(self.radio_button)
+          self.radio_button.setChecked(drivers[section][driver]['selected'])
+
+          if section == 'no_driver':
+            self.no_drv.append(self.radio_button)
+          if section in ('manually_install', 'no_driver') or ('builtin' in drivers[section][driver] and drivers[section][driver]['builtin']):
+            self.radio_button.toggled.connect(partial( self.on_driver_selection_changed, device))
+          else:
+            self.radio_button.toggled.connect(partial( self.on_driver_selection_changed, device, driver))
+          if drivers['manually_installed'] and section != 'manually_installed':
+            self.radio_button.setEnabled(False)
+
+      self.userinterface.box_driver_detail.addLayout(device_box)
+
+    self.userinterface.box_driver_detail.addStretch()
+    self.ui_building = False
+    #self.userinterface.box_driver_detail.show_all()
+    self.set_driver_action_status()
+
+  def update_label_and_icons_from_status(self):
+    '''Update the current label and icon, computing the new device status'''
+
+    for device in self.devices:
+      (overall_status, icon, drivers) = self.gather_device_data(self.devices[device])
+      (driver_status, widget) = self.dynamic_device_status[device]
+
+      pixmap = QIcon.fromTheme(icon).pixmap(QSize(16, 16))
+      driver_status.setPixmap(pixmap)
+      widget.setText("<small>{}</small>".format(overall_status))
+
+  def set_driver_action_status(self):
+    # Update the label in case we end up having some kind of proprietary driver in use.
+    if (os.path.exists('/var/run/reboot-required')):
+      self.userinterface.label_driver_action.setText("You need to restart the computer to complete the driver changes.")
+      self.button_driver_restart.setVisible(True)
+      return
+
+    self.nonfree_drivers = 0
+    for device in self.devices:
+      for pkg_name in self.devices[device]['drivers']:
+        pkg = self.apt_cache[pkg_name]
+        if not self.devices[device]['drivers'][pkg_name]['free'] and pkg.is_installed:
+          self.nonfree_drivers = self.nonfree_drivers + 1
+
+    if self.nonfree_drivers > 0:
+      text = "%s proprietary driver(s) in use." % ( self.nonfree_drivers)
+      self.userinterface.label_driver_action.setText(text)
+    else:
+      self.userinterface.label_driver_action.setText(_("No proprietary drivers are in use."))
