@@ -92,7 +92,7 @@ class TestAptSources(unittest.TestCase):
         apt_pkg.config.set("Dir::Etc::sourcelist",
                            self.sources_list+".testDistribution")
         sources = aptsources.sourceslist.SourcesList()
-        distro = aptsources.distro.get_distro(codename="edgy")
+        distro = aptsources.distro.get_distro(codename="edgy", id="Ubuntu", description="Ubuntu Edgy", release="edgy")
         distro.get_sources(sources)
         # test if all suits of the current distro were detected correctly
         dist_templates = set()
